/**
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2 of
* the License, or (at your option) any later version.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @package    ArmarX::
* @author     Mirko Waechter ( mirko.waechter at kit dot edu)
* @date       2013
* @copyright  http://www.gnu.org/licenses/gpl.txt
*             GNU General Public License
*/

#ifndef _ARMARX_TCPCONTROLUNIT_H
#define _ARMARX_TCPCONTROLUNIT_H

#include <RobotAPI/interface/units/TCPControlUnit.h>
#include <Core/robotstate/remote/ArmarPose.h>
#include <Core/core/services/tasks/PeriodicTask.h>
#include <Core/core/Component.h>

#include <VirtualRobot/IK/DifferentialIK.h>
#include <Core/robotstate/remote/RemoteRobot.h>
#include <Core/interface/observers/ObserverInterface.h>

namespace armarx {


    class TCPControlUnitPropertyDefinitions:
            public ComponentPropertyDefinitions
    {
    public:
        TCPControlUnitPropertyDefinitions(std::string prefix):
            ComponentPropertyDefinitions(prefix)
        {
            defineRequiredProperty<std::string>("KinematicUnitName","Name of the KinematicUnit Proxy");
            defineOptionalProperty<float>("MaxJointVelocity",30.f/180*3.141, "Maximal joint velocity in rad/sec");
            defineRequiredProperty<std::string>("RobotFileName", "Robot file name, e.g. robot_model.xml");
            defineOptionalProperty<int>("CycleTime", 30, "Cycle time of the tcp control in ms");

        }
    };

    class TCPControlUnit :
            virtual public Component,
            virtual public TCPControlUnitInterface

    {
    public:
        TCPControlUnit();


        void onInitComponent();
        void onConnectComponent();
        void onDisconnectComponent();
        void onExitComponent();
        std::string getDefaultName() const{return "TCPControlUnit";}

        // TCPControlUnitInterface interface

        void setCycleTime(Ice::Int milliseconds, const Ice::Current &c = Ice::Current());
        void setTCPVelocity(const std::string &nodeSetName, const std::string &tcpName, const::armarx::FramedVector3BasePtr &translationVelocity, const::armarx::FramedVector3BasePtr &orientationVelocityRPY, const Ice::Current &c = Ice::Current());

        // UnitExecutionManagementInterface interface
        void init(const Ice::Current &c = Ice::Current());
        void start(const Ice::Current &c = Ice::Current());
        void stop(const Ice::Current &c = Ice::Current());
        UnitExecutionState getExecutionState(const Ice::Current &c = Ice::Current());

        // UnitResourceManagementInterface interface
        void request(const Ice::Current &c = Ice::Current());
        void release(const Ice::Current &c = Ice::Current());

        // PropertyUser interface
        PropertyDefinitionsPtr createPropertyDefinitions();

        static Eigen::VectorXf CalcNullspaceJointDeltas(const Eigen::VectorXf &desiredJointDeltas, const Eigen::MatrixXf &jacobian, const Eigen::MatrixXf &jacobianInverse);
        static Eigen::VectorXf CalcJointLimitAvoidanceDeltas(VirtualRobot::RobotNodeSetPtr robotNodeSet, const Eigen::MatrixXf &jacobian, const Eigen::MatrixXf &jacobianInverse, Eigen::VectorXf desiredJointValues = Eigen::VectorXf());
        void calcAndSetVelocities();
    private:
        static void ContinuousAngles(const Eigen::AngleAxisf &oldAngle, Eigen::AngleAxisf &newAngle, double & offset);
        void periodicExec();
        void periodicReport(double tDelta);
        Eigen::VectorXf calcJointVelocities(VirtualRobot::RobotNodeSetPtr robotNodeSet, Eigen::VectorXf tcpDelta, VirtualRobot::RobotNodePtr refFrame, VirtualRobot::IKSolver::CartesianSelection mode);
        Eigen::VectorXf applyMaxJointVelocity(const Eigen::VectorXf &jointVelocity, float maxJointVelocity);

        float maxJointVelocity;
        int cycleTime;
        Eigen::Matrix4f lastTCPPose;

        struct TCPVelocityData
        {
            FramedVector3Ptr translationVelocity;
            FramedVector3Ptr orientationVelocity;
            std::string nodeSetName;
            std::string tcpName;
        };

        typedef std::map<std::string, TCPVelocityData> TCPVelocityDataMap;
        TCPVelocityDataMap tcpVelocitiesMap, localTcpVelocitiesMap;

        typedef std::map<std::string, VirtualRobot::DifferentialIKPtr> DIKMap;
        DIKMap dIKMap, localDIKMap;
//        FramedVector3Map targetTranslationVelocities;
//        FramedVector3Map targetOrientationVelocitiesRPY;

        PeriodicTask<TCPControlUnit>::pointer_type execTask;
        PeriodicTask<TCPControlUnit>::pointer_type topicTask;

        RobotStateComponentInterfacePrx robotStateComponentPrx;
        KinematicUnitInterfacePrx kinematicUnitPrx;
        SharedRobotInterfacePrx remoteRobotPrx;
        VirtualRobot::RobotPtr jointExistenceCheckRobot;
        VirtualRobot::RobotPtr localRobot;
        VirtualRobot::RobotPtr localReportRobot;
        VirtualRobot::RobotPtr localVelReportRobot;
        TCPControlUnitListenerPrx listener;
//        VirtualRobot::RobotNodeSetPtr robotNodeSet;
//        VirtualRobot::DifferentialIKPtr dIK;
        DebugObserverInterfacePrx debugObs;

        bool requested;
        std::map<std::string, double> oriOffsets;


        FramedPoseBaseMap lastTCPPoses;
        IceUtil::Time lastReportTime;
        IceUtil::Time lastTopicReportTime;


        Mutex dataMutex;
        Mutex taskMutex;
        Mutex reportMutex;
        bool calculationRunning;

        std::string topicName;



        // KinematicUnitListener interface
    public:
        void reportControlModeChanged(const NameControlModeMap &, bool, const Ice::Current &);
        void reportJointAngles(const NameValueMap &jointAngles, bool, const Ice::Current &);
        void reportJointVelocities(const NameValueMap &jointVel, bool, const Ice::Current &);
        void reportJointTorques(const NameValueMap &, bool, const Ice::Current &);
        void reportJointCurrents(const NameValueMap &, bool, const Ice::Current &);
        void reportJointMotorTemperatures(const NameValueMap &, bool, const Ice::Current &);
        void reportJointStatuses(const NameStatusMap &, bool, const Ice::Current &);

    };

    class EDifferentialIK : public VirtualRobot::DifferentialIK, virtual public Logging
    {
    public:
        typedef Eigen::VectorXf (EDifferentialIK::*ComputeFunction)(float);

        EDifferentialIK(VirtualRobot::RobotNodeSetPtr rns,VirtualRobot:: RobotNodePtr coordSystem = VirtualRobot::RobotNodePtr(), VirtualRobot::JacobiProvider::InverseJacobiMethod invJacMethod = eSVD);

        VirtualRobot::RobotNodePtr getRefFrame(){return coordSystem;}
        int getJacobianRows(){ return nRows;}

        Eigen::MatrixXf calcFullJacobian();

        void clearGoals();
        void setRefFrame(VirtualRobot::RobotNodePtr coordSystem);

        void setGoal(const Eigen::Matrix4f &goal, VirtualRobot::RobotNodePtr tcp, VirtualRobot::IKSolver::CartesianSelection mode, float tolerancePosition, float toleranceRotation, Eigen::VectorXf tcpWeight);

        void setDOFWeights(Eigen::VectorXf dofWeights);
//        void setTCPWeights(Eigen::VectorXf tcpWeights);
        bool computeSteps(float stepSize = 1.f, float mininumChange = 0.01f, int maxNStep = 50);
        bool computeSteps(Eigen::VectorXf& resultJointDelta, float stepSize = 1.f, float mininumChange = 0.01f, int maxNStep = 50, ComputeFunction computeFunction = NULL);
        Eigen::VectorXf computeStep(float stepSize);
        void applyDOFWeightsToJacobian(Eigen::MatrixXf &Jacobian);
        void applyTCPWeights(VirtualRobot::RobotNodePtr tcp, Eigen::MatrixXf &partJacobian);
        void applyTCPWeights(Eigen::MatrixXf &invJacobian);
        float getWeightedError();
        float getWeightedErrorPosition(VirtualRobot::SceneObjectPtr tcp);
        Eigen::VectorXf computeStepIndependently(float stepSize);
        bool solveIK(float stepSize = 1, float minChange = 0.01f, int maxSteps = 50, bool useAlternativeOnFail = false);
    protected:
        bool adjustDOFWeightsToJointLimits(const Eigen::VectorXf &plannedJointDeltas);

        Eigen::VectorXf dofWeights;
        std::map<VirtualRobot:: SceneObjectPtr, Eigen::VectorXf> tcpWeights;
        Eigen::VectorXf tcpWeightVec;
    };
    typedef boost::shared_ptr<EDifferentialIK> EDifferentialIKPtr;

}

#endif
